#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://github.com/calamares> ===
#
#   Copyright 2014, Rohan Garg <rohan@kde.org>
#   Copyright 2015, Philip Müller <philm@manjaro.org>
#   Copyright 2016, David McKinney <mckinney@subgraph.com>
#   Copyright 2016, Kevin Kofler <kevin.kofler@chello.at>
#   Copyright 2017, Alf Gaida <agaida@siduction.org>
#   Copyright 2017, Adriaan de Groot <groot@kde.org>
#   Copyright 2018, Matthias Klumpp <matthias.klumpp@puri.sm>
#
#   Calamares is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Calamares is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Calamares. If not, see <http://www.gnu.org/licenses/>.

import libcalamares
from libcalamares.utils import check_target_env_call

import os
import shutil
import inspect
import json


CRYPTTAB_HEADER = """# /etc/crypttab: mappings for encrypted partitions.
#
# Each mapped device will be created in /dev/mapper, so your /etc/fstab
# should use the /dev/mapper/<name> paths for encrypted devices.
#
# See crypttab(5) for the supported syntax.
#
# NOTE: Do not list your root (/) partition here, it must be set up
#       beforehand by the initramfs (/etc/mkinitcpio.conf). The same applies
#       to encrypted swap, which should be set up with mkinitcpio-openswap
#       for resume support.
#
# <name>               <device>                         <password> <options>"""


def mkdir_p(path):
    """ Create directory.

    :param path:
    """
    if not os.path.exists(path):
        os.makedirs(path)


def copy_initramfs_hooks(partitions, root_mount_point):
    """
    Copies initramfs hooks so they are picked up by update-initramfs.
    This unlocks the encrypted partition by default, and allows the user to
    reencrypt it with a different password.

    :param partitions:
    :param root_mount_point:
    """
    encrypt_hook = False

    for partition in partitions:
        if partition["mountPoint"] == "/" and "luksMapperName" in partition:
            encrypt_hook = True

    if encrypt_hook:
        target = "{!s}/usr/share/initramfs-tools/hooks/bypass_encrypt_hook".format(root_mount_point)

        # Find where this module is installed
        _filename = inspect.getframeinfo(inspect.currentframe()).filename
        _path = os.path.dirname(os.path.abspath(_filename))

        shutil.copy2(
            os.path.join(_path, "bypass_encrypt_hook"),
            target)
        os.chmod(target, 0o755)


def get_device_id_alias(device):
    from glob import glob
    all_disk_paths = glob('/dev/disk/by-id/*')

    device_id_alias = device
    for d in all_disk_paths:
        # resolve alias links to direct /dev nodes
        dev_path = os.path.realpath(d)
        if dev_path == device:
            device_id_alias = d
            break

    return device_id_alias


def generate_crypttab_line_info(partition, crypttab_options):
    """ Generates information for each crypttab entry. """
    if "luksMapperName" not in partition or "luksUuid" not in partition:
        return None

    mapper_name = partition["luksMapperName"]
    luks_uuid = partition["luksUuid"]
    if not mapper_name or not luks_uuid:
        return None

    if partition["fs"] == "linuxswap":
        # we always re-encrypt swap partitions with a random password on boot
        # this makes suspend-to-disk impossible, but improves security slightly
        return dict(
            name=mapper_name,
            device=get_device_id_alias(partition["device"]),
            password="/dev/urandom",
            options="swap,cipher=aes-xts-plain64,size=256,noearly",
        )

    return dict(
        name=mapper_name,
        device="UUID=" + luks_uuid,
        password="/crypto_keyfile.bin",
        options=crypttab_options,
    )


def print_crypttab_line(dct, file=None):
    """ Prints line to '/etc/crypttab' file. """
    line = "{:21} {:<45} {} {}".format(dct["name"],
                                       dct["device"],
                                       dct["password"],
                                       dct["options"],
                                       )

    print(line, file=file)


def generate_crypttab(partitions, root_mount_point, crypttab_options):
    """ Create crypttab. """
    mkdir_p(os.path.join(root_mount_point, "etc"))
    crypttab_path = os.path.join(root_mount_point, "etc", "crypttab")

    with open(crypttab_path, "w") as crypttab_file:
        print(CRYPTTAB_HEADER, file=crypttab_file)

        for partition in partitions:
            dct = generate_crypttab_line_info(partition, crypttab_options)

            if dct:
                print_crypttab_line(dct, file=crypttab_file)


def remove_disk_passwords_store_state(partitions, root_mount_point):
    """ We assume that the luksbootkeyfile module has been run priot to this
    module, so we can actually create keyfile-only partitions """

    def device_entry_from_partition(p):
        entry = {}
        entry["device"] = p["device"]
        entry["luksPassphrase"] = p.get("luksPassphrase", None)
        entry["luksMapperName"] = p["luksMapperName"]
        entry["luksUuid"] = p["luksUuid"]
        entry["fs"] = p["fs"]
        return entry

    luks_root_device = {}
    additional_luks_devices = []
    for partition in partitions:
        if partition["mountPoint"] == "/" and "luksMapperName" in partition:
            luks_root_device = device_entry_from_partition(partition)
        elif "luksMapperName" in partition and\
             (partition["mountPoint"] or partition["fs"] == "linuxswap"):
            additional_luks_devices.append(device_entry_from_partition(partition))

    if not luks_root_device:
        return True, None
    if not luks_root_device["luksPassphrase"]:
        return False, ("Encrypted rootfs setup error",
                       "Rootfs partition {!s} is LUKS but no passphrase found."
                       .format(luks_root_device))

    # remove passwords
    check_target_env_call(["cryptsetup",
                           "luksRemoveKey",
                           luks_root_device["device"]],
                          luks_root_device["luksPassphrase"],
                          15)  # timeout 15s

    additional_luks_devices_info = []
    for device in additional_luks_devices:
        luks_passphrase = device.pop("luksPassphrase")  # drop potential passphrase, so we don't write it to disk later
        additional_luks_devices_info.append(device)
        if not luks_passphrase:
            continue

        check_target_env_call(["cryptsetup",
                               "luksRemoveKey",
                               device["device"]],
                              luks_passphrase,
                              15)  # timeout 15s

    # we don't want to expose an already removed passphrase in clear text
    luks_root_device.pop("luksPassphrase")

    encrypted_state_file_path = os.path.join(root_mount_point, "encrypted_partitions.json")
    with open(encrypted_state_file_path, "w") as encrypted_state_file:
        json_data = json.dumps({'rootDevice': luks_root_device, 'additionalDevices': additional_luks_devices_info})
        print(json_data, file=encrypted_state_file)

    return True, None


def run():
    """
    Calls routine with given parameters to configure initramfs
    and /etc/crypttab to allow passwordless decryption on the first boot
    of the system.

    :return:
    """
    partitions = libcalamares.globalstorage.value("partitions")
    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")

    res, msg = remove_disk_passwords_store_state(partitions, root_mount_point)
    if not res:
        return msg
    copy_initramfs_hooks(partitions, root_mount_point)
    generate_crypttab(partitions, root_mount_point, 'luks,keyscript=/bin/cat')

    return None
